/*
   checksum.c -- PIKT checksum routines

   $Header: /var/cvs/PIKT/pikt/src/lib/checksum.c,v 1.16 2007/01/18 13:13:09 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2007 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   Portions of this file are modifications of code originally written by
   Kayvan Aghaiepour, David MacKenzie, Q. Frank Xia, and L. Peter Deutsch
   (see additional attributions below).

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "common.h"

#include <math.h>

#define PADDING 16	/* extra bytes mymalloc()'ed just to be safe */

char *compact(char *);
char *syscksumfil(const char *, const char *);
char *nullsumfil(const char *);
char *bsdsumfil(const char *);
char *sysvsumfil(const char *);
char *cksumfil(const char *);
char *mdxsumfil(const char *, int);
char *nullsumstr(const char *);
char *bsdsumstr(const char *);
char *sysvsumstr(const char *);
char *cksumstr(const char *);
char *mdxsumstr(const char *, int);

/* for a file, return a string "<checksum> <filesize> <filename>";
   checksumfil(-3, ...) returns the output of the system cksum command;
   checksumfil(-2, ...) returns the output of the system sum command;
   checksumfil(-1, ...) returns the output of the system 'sum -r' command;
   checksumfil( 0, ...) returns the null checksum;
   checksumfil( 1, ...) returns the BSD-style sum computed internally;
   checksumfil( 2, ...) returns the SysV-style sum computed internally;
   checksumfil( 3, ...) returns the POSIX cksum computed internally;
   checksumfil( 4, ...) returns the MD4 checksum;
   checksumfil( 5, ...) returns the MD5 checksum;
   this is all subject to change; in particular, the checksums assigned to
   the different levels might change, and more levels might be added to
   compute, for example, these additional checksums: SNEFRU, HAVAL, SHA
 */
char *
checksumfil (int level, signed char *file)
{
	static char fnnam[] = "checksumfil";
	char *cmd;
	char *r;

	switch (level) {
		case -3: /* default system cksum command checksum */
			cmd = (char *) mymalloc(strlen(cksumcmd)+strlen(file)+2, fnnam);
			sprintf(cmd, "%s %s", cksumcmd, file);
			r = compact(syscksumfil(file, cmd));
			free(cmd);
			break;
		case -2: /* default system sum command checksum */
			cmd = (char *) mymalloc(strlen(sumcmd)+strlen(file)+2, fnnam);
			sprintf(cmd, "%s %s", sumcmd, file);
			r = compact(syscksumfil(file, cmd));
			free(cmd);
			break;
		case -1: /* default system 'sum -r' command checksum */
			cmd = (char *) mymalloc(strlen(sumcmd)+strlen(file)+5, fnnam);
			sprintf(cmd, "%s -r %s", sumcmd, file);
			r = compact(syscksumfil(file, cmd));
			free(cmd);
			break;
		case 0: /* null checksum */
			r = nullsumfil(file);
			break;
		case 1: /* bsd sum */
			r = bsdsumfil(file);
			break;
		case 2: /* sysv sum */
			r = sysvsumfil(file);
			break;
		case 3: /* posix cksum */
			r = cksumfil(file);
			break;
		case 4: /* md4 checksum */
			r = mdxsumfil(file, 4);
			break;
		case 5: /* md5 checksum */
			r = mdxsumfil(file, 5);
			break;
		default:
			r = mystrdup(errstr, fnnam);
			break;
	}
	return r;
}

/* for a string, return a string "<checksum>";
   checksumstr( 0, ...) returns the null checksum;
   checksumstr( 1, ...) returns the BSD-style sum;
   checksumstr( 2, ...) returns the SysV-style sum;
   checksumstr( 3, ...) returns the POSIX cksum;
   checksumstr( 4, ...) returns the MD4 checksum;
   checksumstr( 5, ...) returns the MD5 checksum;
   this is all subject to change; in particular, the checksums assigned to
   the different levels might change, and more levels might be added to
   compute, for example, these additional checksums: SNEFRU, HAVAL, SHA
 */
char *
checksumstr (int level, signed char *str)
{
	static char fnnam[] = "checksumstr";
	char *r;

	switch (level) {
		case 0: /* null checksum */
			r = nullsumstr(str);
			break;
		case 1: /* bsd sum */
			r = bsdsumstr(str);
			break;
		case 2: /* sysv sum */
			r = sysvsumstr(str);
			break;
		case 3: /* posix cksum */
			r = cksumstr(str);
			break;
		case 4: /* md4 checksum */
			r = mdxsumstr(str, 4);
			break;
		case 5: /* md5 checksum */
			r = mdxsumstr(str, 5);
			break;
		default:
			r = mystrdup(errstr, fnnam);
			break;
	}
	return r;
}

/* compare two checksum strings, terminating upon reaching the file paths
   (beginning at the second space of the checksum string), which assuredly
   will differ; return 1 if the checksum and filesize differ, 0 if they match;

   the function assumes checksum strings of the form:

   43d2dbd8e8d597663a963208f0a43f8d 1375 /etc/syslog.conf

   that is, checksum, followed by file size, followed by file path, each
   separated by a single whitespace (like what the PIKT checksum() returns)
 */
int
csstrcmp (char *cs1, char *cs2)
{
	register int i;
	register int spcnt;	/* spaces count */
	register char c1, c2;

	spcnt = 0;
	c1 = c2 = '\0';
	i = 0;
	while ((c1 = *(cs1+i)) && (c2 = *(cs2+i)) && (spcnt < 2)) {
		if (c1 != c2)
			return 1;
		if (isspace((int)c1) || isspace((int)c2))
			spcnt++;
		i++;
	}
	if ((c1 && ! c2) || (! c1 && c2) || (spcnt < 2)) {	/* equivalent
								   to: c1 != c2
								 */
		return 1;
	} else {
		return 0;
	}
}

/* return the first output line for the given system checksum command
 */
char *
syscksumfil (const char *file, const char *cmd)
{
	static char fnnam[] = "syscksumfil";
	struct stat fbuf;
	FILE *infile;
	char buf[MAXBUF+1];
	char *r;

	if (stat(file, &fbuf) != 0) {
		if (isverboselog)
			logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "stat() error for %s", file);
		r = mystrdup(err, fnnam);
		return r;
	}
	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "command \"%s\"", cmd);
	if ((infile = popen(cmd, "r")) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_PROC_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "popen() error for \"%s\"", cmd);
		r = mystrdup(err, fnnam);
		return r;
	}
	if (fgets(buf, MAXBUF, infile) == NULL) {
		if (ferror(infile)) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_FGETS_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fgets() error for \"%s\"", cmd);
			/* clearerr(infile); */
			r = mystrdup(err, fnnam);
		} else {
			r = mystrdup(nullstring, fnnam);
		}
		if (pclose(infile) == -1)
			logreturn(PIKTLOG, LOG_ERR, ERRID_PROC_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "pclose() error for \"%s\"", cmd);
		return r;
	}
	if (pclose(infile) == ERROR)
		logreturn(PIKTLOG, LOG_ERR, ERRID_PROC_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "pclose() error for \"%s\"", cmd);
	r = mystrdup(buf, fnnam);
	*strchr(r, '\n') = '\0';
	return r;
}

/* compact the given string, replacing consecutive whitespaces with a single
   space
 */
char *
compact (char *r)
{
	static char fnnam[] = "compact";
	register int i, j;
	register signed char c;
	register int wasspace;
	char *s;

	s = mystrdup(r, fnnam);
	wasspace = FALSE;
	i = j = 0;
	while ((c = *(s+i))) {
		if (isspace((int)c)) {
			if (! wasspace) {
				*(r+j) = ' ';
				j++;
			}
			wasspace = TRUE;
		} else {
			wasspace = FALSE;
			*(r+j) = c;
			j++;
		}
		i++;
	}
	*(r+j) = '\0';
	free(s);
	return r;
}

/*****************************************************************************/

/* return the null checksum (0), also the number of bytes, for the given file
 */
char *
nullsumfil (const char *file)
{
	static char fnnam[] = "nullsumfil";
	struct stat fbuf;
	register long total_bytes;
	char *r;

	if (stat(file, &fbuf) != 0) {
		if (isverboselog)
			logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "stat() error for %s", file);
		total_bytes = ERROR;
	} else {
		total_bytes = fbuf.st_size;
	}
	if (total_bytes == ERROR)
		r = mystrdup(errstr, fnnam);
	else {
		r = mymalloc(strlen(file)+14+PADDING, fnnam);
		sprintf(r, "0 %ld %s", total_bytes, file);
	}
	return r;
}

/* return the null checksum (0) for the given string
 */
char *
nullsumstr (const char *str)
{
	char *r;

	r = mystrdup("0", "nullsumstr");
	return r;
}

/*****************************************************************************/

/* calculate the BSD-like rotated checksum; the checksum varies depending
   on sizeof(int)

   Copyright (C) 86, 89, 91, 95, 1996 Free Software Foundation, Inc.
   Distributed under the GNU GPL.
   Written by Kayvan Aghaiepour and David MacKenzie.
   Modified by Robert Osterlund for use with PIKT.
 */

/* right-rotate 32-bit integer variable c */
#define ROTATE_RIGHT(c) if ((c) & 01) (c) = ((c) >>1) + 0x8000; else (c) >>= 1;

/* calculate the BSD-like rotated checksum, also the size in 512-bye
   blocks, for the given file
 */
char *
bsdsumfil (const char *file)
{
	static char fnnam[] = "bsdsumfil";
	register FILE *fp;
	register unsigned long checksum;	/* The checksum mod 2^16. */
	register long total_bytes;		/* The number of bytes. */
	register int ch;			/* Each character read. */
	char *r;

	checksum = 0;
	total_bytes = 0;
	if ((fp = fopen(file, "r")) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", file);
		r = mystrdup(errstr, fnnam);
		return r;
	}
	while ((ch = getc(fp)) != EOF) {
		total_bytes++;
		ROTATE_RIGHT(checksum);
		checksum += ch;
		checksum &= 0xffff;	/* Keep it within bounds. */
	}
	if (ferror(fp) || fclose(fp)) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "getc() or fclose() error for %s", file);
		r = mystrdup(errstr, fnnam);
		return r;
	}
	r = mymalloc(strlen(file)+13+PADDING, fnnam);
	sprintf(r, "%05lu %ld %s", checksum, (total_bytes+512-1)/512, file);
	return r;
}

/* calculate the BSD-like rotated checksum for the given string
 */
char *
bsdsumstr (const char *str)
{
	static char fnnam[] = "bsdsumstr";
	register unsigned long checksum;	/* The checksum mod 2^16. */
	register int i;				/* str offset */
	register int ch;			/* Each character read. */
	char *r;

	checksum = 0;
	i = 0;
	while ((ch = *(str+i))) {
		ROTATE_RIGHT(checksum);
		checksum += ch;
		checksum &= 0xffff;	/* Keep it within bounds. */
		i++;
	}
	r = mymalloc(6, fnnam);
	sprintf(r, "%05lu", checksum);
	return r;
}

/*****************************************************************************/

/* calculate the SysV checksum

   Copyright (C) 86, 89, 91, 95, 1996 Free Software Foundation, Inc.
   Distributed under the GNU GPL.
   Written by Kayvan Aghaiepour and David MacKenzie.
   Modified by Robert Osterlund for use with PIKT.
 */

/* int safe_read(int, char *, int); */

/* calculate the SysV checksum, also the size in 512-byte blocks, for the
   given file
 */
char *
sysvsumfil (const char *file)
{
	static char fnnam[] = "sysvsumfil";
	int fd;
	unsigned char buf[8192];
	register int bytes_read;
	register unsigned long checksum;
	long total_bytes;
	char *r;

	checksum = 0;
	total_bytes = 0;
	if ((fd = open(file, O_RDONLY)) == -1) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "open() error for %s", file);
		r = mystrdup(errstr, fnnam);
		return r;
	}
	/* safe_read() was used in the original implementation, but why
	   bother here when we don't elsewhere?
	 */
	while ((bytes_read = /* safe_read */ read(fd, buf, sizeof buf)) > 0) {
		register int i;
		for (i=0; i<bytes_read; i++)
			checksum += buf[i];
		total_bytes += bytes_read;
	}
	if ((bytes_read < 0) || (close(fd) == -1)) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "safe_read() or close() error for %s", file);
		r = mystrdup(errstr, fnnam);
		return r;
	}
	r = mymalloc(strlen(file)+13+PADDING, fnnam);
	sprintf(r, "%lu %ld %s", checksum % 0xffff, (total_bytes+512-1)/512, file);
	return r;
}

/* calculate the SysV checksum for the given string
 */
char *
sysvsumstr (const char *str)
{
	static char fnnam[] = "sysvsumstr";
	register unsigned long checksum;
	register int i;
	register int ch;
	char *r;

	checksum = 0;
	i = 0;
	while ((ch = *(str+i))) {
		checksum += ch;
		i++;
	}
	r = mymalloc(64, fnnam);
	sprintf(r, "%lu", checksum % 0xffff);
	return r;
}

/* safe_read(), an interface to read that retries after interrupts;
   read LEN bytes at PTR from descriptor DESC, retrying if interrupted;
   return the actual number of bytes read, zero for EOF, or negative
   for an error.

   Copyright (C) 1993, 1994 Free Software Foundation, Inc.
   Modified by Robert Osterlund for use with PIKT.
 */
/*
int
safe_read (int desc, char *ptr, int len)
{
	int n_chars;

	if (len <= 0)
		return len;
	#ifdef EINTR
	do {
		n_chars = read(desc, ptr, len);
	} while ((n_chars < 0) && (errno == EINTR));
	#else
		n_chars = read(desc, ptr, len);
	#endif
	return n_chars;
}
 */

/*****************************************************************************/

/* calculate and print POSIX.2 checksum

   This software is compatible with neither the System V nor the BSD
   `sum' program.  It is supposed to conform to P1003.2/D11.2,
   except foreign language interface (4.9.5.3 of P1003.2/D11.2) support.
   Any inconsistency with the standard except 4.9.5.3 is a bug.

   Copyright (C) 92, 95, 1996 Free Software Foundation, Inc.
   Written by Q. Frank Xia, qx@math.columbia.edu.
   Cosmetic changes and reorganization by David MacKenzie, djm@gnu.ai.mit.edu.
   Modified by Robert Osterlund for use with PIKT.
*/

/* number of bytes to read at once */
#define BUFLEN (1 << 16)

static unsigned long const crctab[256] =
{
  0x0,
  0x04C11DB7, 0x09823B6E, 0x0D4326D9, 0x130476DC, 0x17C56B6B,
  0x1A864DB2, 0x1E475005, 0x2608EDB8, 0x22C9F00F, 0x2F8AD6D6,
  0x2B4BCB61, 0x350C9B64, 0x31CD86D3, 0x3C8EA00A, 0x384FBDBD,
  0x4C11DB70, 0x48D0C6C7, 0x4593E01E, 0x4152FDA9, 0x5F15ADAC,
  0x5BD4B01B, 0x569796C2, 0x52568B75, 0x6A1936C8, 0x6ED82B7F,
  0x639B0DA6, 0x675A1011, 0x791D4014, 0x7DDC5DA3, 0x709F7B7A,
  0x745E66CD, 0x9823B6E0, 0x9CE2AB57, 0x91A18D8E, 0x95609039,
  0x8B27C03C, 0x8FE6DD8B, 0x82A5FB52, 0x8664E6E5, 0xBE2B5B58,
  0xBAEA46EF, 0xB7A96036, 0xB3687D81, 0xAD2F2D84, 0xA9EE3033,
  0xA4AD16EA, 0xA06C0B5D, 0xD4326D90, 0xD0F37027, 0xDDB056FE,
  0xD9714B49, 0xC7361B4C, 0xC3F706FB, 0xCEB42022, 0xCA753D95,
  0xF23A8028, 0xF6FB9D9F, 0xFBB8BB46, 0xFF79A6F1, 0xE13EF6F4,
  0xE5FFEB43, 0xE8BCCD9A, 0xEC7DD02D, 0x34867077, 0x30476DC0,
  0x3D044B19, 0x39C556AE, 0x278206AB, 0x23431B1C, 0x2E003DC5,
  0x2AC12072, 0x128E9DCF, 0x164F8078, 0x1B0CA6A1, 0x1FCDBB16,
  0x018AEB13, 0x054BF6A4, 0x0808D07D, 0x0CC9CDCA, 0x7897AB07,
  0x7C56B6B0, 0x71159069, 0x75D48DDE, 0x6B93DDDB, 0x6F52C06C,
  0x6211E6B5, 0x66D0FB02, 0x5E9F46BF, 0x5A5E5B08, 0x571D7DD1,
  0x53DC6066, 0x4D9B3063, 0x495A2DD4, 0x44190B0D, 0x40D816BA,
  0xACA5C697, 0xA864DB20, 0xA527FDF9, 0xA1E6E04E, 0xBFA1B04B,
  0xBB60ADFC, 0xB6238B25, 0xB2E29692, 0x8AAD2B2F, 0x8E6C3698,
  0x832F1041, 0x87EE0DF6, 0x99A95DF3, 0x9D684044, 0x902B669D,
  0x94EA7B2A, 0xE0B41DE7, 0xE4750050, 0xE9362689, 0xEDF73B3E,
  0xF3B06B3B, 0xF771768C, 0xFA325055, 0xFEF34DE2, 0xC6BCF05F,
  0xC27DEDE8, 0xCF3ECB31, 0xCBFFD686, 0xD5B88683, 0xD1799B34,
  0xDC3ABDED, 0xD8FBA05A, 0x690CE0EE, 0x6DCDFD59, 0x608EDB80,
  0x644FC637, 0x7A089632, 0x7EC98B85, 0x738AAD5C, 0x774BB0EB,
  0x4F040D56, 0x4BC510E1, 0x46863638, 0x42472B8F, 0x5C007B8A,
  0x58C1663D, 0x558240E4, 0x51435D53, 0x251D3B9E, 0x21DC2629,
  0x2C9F00F0, 0x285E1D47, 0x36194D42, 0x32D850F5, 0x3F9B762C,
  0x3B5A6B9B, 0x0315D626, 0x07D4CB91, 0x0A97ED48, 0x0E56F0FF,
  0x1011A0FA, 0x14D0BD4D, 0x19939B94, 0x1D528623, 0xF12F560E,
  0xF5EE4BB9, 0xF8AD6D60, 0xFC6C70D7, 0xE22B20D2, 0xE6EA3D65,
  0xEBA91BBC, 0xEF68060B, 0xD727BBB6, 0xD3E6A601, 0xDEA580D8,
  0xDA649D6F, 0xC423CD6A, 0xC0E2D0DD, 0xCDA1F604, 0xC960EBB3,
  0xBD3E8D7E, 0xB9FF90C9, 0xB4BCB610, 0xB07DABA7, 0xAE3AFBA2,
  0xAAFBE615, 0xA7B8C0CC, 0xA379DD7B, 0x9B3660C6, 0x9FF77D71,
  0x92B45BA8, 0x9675461F, 0x8832161A, 0x8CF30BAD, 0x81B02D74,
  0x857130C3, 0x5D8A9099, 0x594B8D2E, 0x5408ABF7, 0x50C9B640,
  0x4E8EE645, 0x4A4FFBF2, 0x470CDD2B, 0x43CDC09C, 0x7B827D21,
  0x7F436096, 0x7200464F, 0x76C15BF8, 0x68860BFD, 0x6C47164A,
  0x61043093, 0x65C52D24, 0x119B4BE9, 0x155A565E, 0x18197087,
  0x1CD86D30, 0x029F3D35, 0x065E2082, 0x0B1D065B, 0x0FDC1BEC,
  0x3793A651, 0x3352BBE6, 0x3E119D3F, 0x3AD08088, 0x2497D08D,
  0x2056CD3A, 0x2D15EBE3, 0x29D4F654, 0xC5A92679, 0xC1683BCE,
  0xCC2B1D17, 0xC8EA00A0, 0xD6AD50A5, 0xD26C4D12, 0xDF2F6BCB,
  0xDBEE767C, 0xE3A1CBC1, 0xE760D676, 0xEA23F0AF, 0xEEE2ED18,
  0xF0A5BD1D, 0xF464A0AA, 0xF9278673, 0xFDE69BC4, 0x89B8FD09,
  0x8D79E0BE, 0x803AC667, 0x84FBDBD0, 0x9ABC8BD5, 0x9E7D9662,
  0x933EB0BB, 0x97FFAD0C, 0xAFB010B1, 0xAB710D06, 0xA6322BDF,
  0xA2F33668, 0xBCB4666D, 0xB8757BDA, 0xB5365D03, 0xB1F740B4
};

/* calculate and print POSIX.2 checksum and size for the given file
 */
char *
cksumfil (const char *file)
{
	static char fnnam[] = "cksumfil";
	unsigned char buf[BUFLEN];
	unsigned long crc;
	long length;
	long bytes_read;
	register FILE *fp;
	char *r;

	crc = 0;
	length = 0;
	if ((fp = fopen(file, "r")) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", file);
		r = mystrdup(errstr, fnnam);
		return r;
	}
	while ((bytes_read = fread(buf, 1, BUFLEN, fp)) > 0) {
		unsigned char *cp = buf;
		length += bytes_read;
		while (bytes_read--)
			crc = (crc << 8) ^ crctab[((crc >> 24) ^ *(cp++)) & 0xFF];
	}
	if (ferror(fp) || fclose(fp)) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fread() or fclose() error for %s", file);
		r = mystrdup(errstr, fnnam);
		return r;
	}
	bytes_read = length;
	while (bytes_read > 0) {
		crc = (crc << 8) ^ crctab[((crc >> 24) ^ bytes_read) & 0xFF];
		bytes_read >>= 8;
	}
	crc = ~crc & 0xFFFFFFFF;
	r = mymalloc(strlen(file)+18+PADDING, fnnam);
	sprintf(r, "%lu %ld %s", crc, length, file);
	return r;
}

/* calculate and print POSIX.2 checksum for the given string
 */
char *
cksumstr (const char *str)
{
	static char fnnam[] = "cksumstr";
	unsigned char buf[BUFLEN];
	unsigned long crc;
	long length;
	long bytes_read;
	long i;
	char *r;

	crc = 0;
	length = 0;
	bytes_read = 0;
	i = 0;
	while ((buf[bytes_read] = (unsigned char) *(str+i))) {
		bytes_read++;
		i++;
		if ((! *(str+i)) || (bytes_read == BUFLEN)) {
			unsigned char *cp = buf;
			length += bytes_read;
			while (bytes_read--)
				crc = (crc << 8) ^ crctab[((crc >> 24) ^ *(cp++)) & 0xFF];
			bytes_read = 0;
		}
	}
	bytes_read = length;
	while (bytes_read > 0) {
		crc = (crc << 8) ^ crctab[((crc >> 24) ^ bytes_read) & 0xFF];
		bytes_read >>= 8;
	}
	crc = ~crc & 0xFFFFFFFF;
	r = mymalloc(64, fnnam);
	sprintf(r, "%lu", crc);
	return r;
}

/*****************************************************************************/

/*
  Copyright (C) 1999 Aladdin Enterprises.  All rights reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  L. Peter Deutsch
  ghost@aladdin.com

 */

/*
   Independent implementation of MD5 (RFC 1321).

   This code implements the MD5 Algorithm defined in RFC 1321.
   It is derived directly from the text of the RFC and not from the
   reference implementation.

   The original and principal author of md5.h is L. Peter Deutsch
   <ghost@aladdin.com>.  Other authors are noted in the change history
   that follows (in reverse chronological order):

   1999-11-04 lpd Edited comments slightly for automatic TOC extraction.
   1999-10-18 lpd Fixed typo in header comment (ansi2knr rather than md5);
	added conditionalization for C++ compilation from Martin
	Purschke <purschke@bnl.gov>.
   1999-05-03 lpd Original version.

   Modified by Robert Osterlund for use with PIKT.
 */

typedef unsigned char mdx_byte_t;	/* 8-bit byte */
typedef unsigned int mdx_word_t;	/* 32-bit word */

/* Define the state of the MDx Algorithm. */
typedef struct mdx_state_s {
	mdx_word_t count[2];	/* message length in bits, lsw first */
	mdx_word_t abcd[4];	/* digest buffer */
	mdx_byte_t buf[64];	/* accumulate block */
} mdx_state_t;

/* Process the latest string of data. */
void md4_process(mdx_state_t *pms, const mdx_byte_t *data /*[64]*/);
void md5_process(mdx_state_t *pms, const mdx_byte_t *data /*[64]*/);

/* Initialize the algorithm. */
void mdx_init(mdx_state_t *pms, int type);

/* Append a string to the message. */
void mdx_append(mdx_state_t *pms, const mdx_byte_t *data, int nbytes, int type);

/* Finish the message and return the digest. */
void mdx_finish(mdx_state_t *pms, mdx_byte_t digest[16], int type);

/*
 * T values needed for MD5
 */
#define T1 0xd76aa478
#define T2 0xe8c7b756
#define T3 0x242070db
#define T4 0xc1bdceee
#define T5 0xf57c0faf
#define T6 0x4787c62a
#define T7 0xa8304613
#define T8 0xfd469501
#define T9 0x698098d8
#define T10 0x8b44f7af
#define T11 0xffff5bb1
#define T12 0x895cd7be
#define T13 0x6b901122
#define T14 0xfd987193
#define T15 0xa679438e
#define T16 0x49b40821
#define T17 0xf61e2562
#define T18 0xc040b340
#define T19 0x265e5a51
#define T20 0xe9b6c7aa
#define T21 0xd62f105d
#define T22 0x02441453
#define T23 0xd8a1e681
#define T24 0xe7d3fbc8
#define T25 0x21e1cde6
#define T26 0xc33707d6
#define T27 0xf4d50d87
#define T28 0x455a14ed
#define T29 0xa9e3e905
#define T30 0xfcefa3f8
#define T31 0x676f02d9
#define T32 0x8d2a4c8a
#define T33 0xfffa3942
#define T34 0x8771f681
#define T35 0x6d9d6122
#define T36 0xfde5380c
#define T37 0xa4beea44
#define T38 0x4bdecfa9
#define T39 0xf6bb4b60
#define T40 0xbebfbc70
#define T41 0x289b7ec6
#define T42 0xeaa127fa
#define T43 0xd4ef3085
#define T44 0x04881d05
#define T45 0xd9d4d039
#define T46 0xe6db99e5
#define T47 0x1fa27cf8
#define T48 0xc4ac5665
#define T49 0xf4292244
#define T50 0x432aff97
#define T51 0xab9423a7
#define T52 0xfc93a039
#define T53 0x655b59c3
#define T54 0x8f0ccc92
#define T55 0xffeff47d
#define T56 0x85845dd1
#define T57 0x6fa87e4f
#define T58 0xfe2ce6e0
#define T59 0xa3014314
#define T60 0x4e0811a1
#define T61 0xf7537e82
#define T62 0xbd3af235
#define T63 0x2ad7d2bb
#define T64 0xeb86d391

void
mdx_init(mdx_state_t *pms, int type)
{
	/* note: type is unused */

	pms->count[0] = pms->count[1] = 0;
	pms->abcd[0] = 0x67452301;
	pms->abcd[1] = 0xefcdab89;
	pms->abcd[2] = 0x98badcfe;
	pms->abcd[3] = 0x10325476;
}

void
mdx_append(mdx_state_t *pms, const mdx_byte_t *data, int nbytes, int type)
{
	const mdx_byte_t *p = data;
	int left = nbytes;
	int offset = (pms->count[0] >> 3) & 63;
	mdx_word_t nbits = (mdx_word_t)(nbytes << 3);

	if (nbytes <= 0)
		return;

	/* Update the message length. */
	pms->count[1] += nbytes >> 29;
	pms->count[0] += nbits;
	if (pms->count[0] < nbits)
		pms->count[1]++;

	/* Process an initial partial block. */
	if (offset) {
		int copy = (offset + nbytes > 64 ? 64 - offset : nbytes);
		memcpy(pms->buf + offset, p, copy);
		if (offset + copy < 64)
			return;
		p += copy;
		left -= copy;
		if (type == 4)
			md4_process(pms, pms->buf);
		else if (type == 5)
			md5_process(pms, pms->buf);
	}

	/* Process full blocks. */
	for (; left>=64; p += 64, left -= 64)
		if (type == 4)
			md4_process(pms, p);
		else if (type == 5)
			md5_process(pms, p);

	/* Process a final partial block. */
	if (left)
		memcpy(pms->buf, p, left);
}

void
mdx_finish(mdx_state_t *pms, mdx_byte_t digest[16], int type)
{
	static const mdx_byte_t pad[64] = {
		0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
	};
	mdx_byte_t data[8];
	int i;

	/* Save the length before padding. */
	for (i=0; i<8; ++i)
		data[i] = (mdx_byte_t)(pms->count[i >> 2] >> ((i & 3) << 3));
	/* Pad to 56 bytes mod 64. */
	mdx_append(pms, pad, ((55 - (pms->count[0] >> 3)) & 63) + 1, type);
	/* Append the length. */
	mdx_append(pms, data, 8, type);
	for (i=0; i<16; ++i)
		digest[i] = (mdx_byte_t)(pms->abcd[i >> 2] >> ((i & 3) << 3));
}

void
md5_process(mdx_state_t *pms, const mdx_byte_t *data /*[64]*/)
{
	mdx_word_t
		a = pms->abcd[0], b = pms->abcd[1],
		c = pms->abcd[2], d = pms->abcd[3];
	mdx_word_t t;

#ifndef ARCH_IS_BIG_ENDIAN
#  define ARCH_IS_BIG_ENDIAN 1	/* slower, default implementation */
#endif

#if ARCH_IS_BIG_ENDIAN

	/*
	 * On big-endian machines, we must arrange the bytes in the right
	 * order.  (This also works on machines of unknown byte order.)
	 */
	mdx_word_t X[16];
	const mdx_byte_t *xp = data;
	int i;

	for (i=0; i<16; ++i, xp += 4)
		X[i] = xp[0] + (xp[1] << 8) + (xp[2] << 16) + (xp[3] << 24);

#else  /* ! ARCH_IS_BIG_ENDIAN */

	/*
	 * On little-endian machines, we can process properly aligned data
	 * without copying it.
	 */
	mdx_word_t xbuf[16];
	const mdx_word_t *X;

	if (! ((data - (const mdx_byte_t *)0) & 3)) {
		/* data are properly aligned */
		X = (const mdx_word_t *)data;
	} else {
		/* not aligned */
		memcpy(xbuf, data, 64);
		X = xbuf;
	}
#endif

#define ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32 - (n))))

	/* Round 1. */
	/* Let [abcd k s i] denote the operation
	a = b + ((a + F(b,c,d) + X[k] + T[i]) <<< s). */
#define F(x, y, z) (((x) & (y)) | (~(x) & (z)))
#define S(a, b, c, d, k, s, Ti)\
	t = a + F(b,c,d) + X[k] + Ti;\
	a = ROTATE_LEFT(t, s) + b
	/* Do the following 16 operations. */
	S(a, b, c, d,  0,  7,  T1);
	S(d, a, b, c,  1, 12,  T2);
	S(c, d, a, b,  2, 17,  T3);
	S(b, c, d, a,  3, 22,  T4);
	S(a, b, c, d,  4,  7,  T5);
	S(d, a, b, c,  5, 12,  T6);
	S(c, d, a, b,  6, 17,  T7);
	S(b, c, d, a,  7, 22,  T8);
	S(a, b, c, d,  8,  7,  T9);
	S(d, a, b, c,  9, 12, T10);
	S(c, d, a, b, 10, 17, T11);
	S(b, c, d, a, 11, 22, T12);
	S(a, b, c, d, 12,  7, T13);
	S(d, a, b, c, 13, 12, T14);
	S(c, d, a, b, 14, 17, T15);
	S(b, c, d, a, 15, 22, T16);
#undef S
#undef F

	 /* Round 2. */
	 /* Let [abcd k s i] denote the operation
          a = b + ((a + G(b,c,d) + X[k] + T[i]) <<< s). */
#define G(x, y, z) (((x) & (z)) | ((y) & ~(z)))
#define S(a, b, c, d, k, s, Ti)\
	t = a + G(b,c,d) + X[k] + Ti;\
	a = ROTATE_LEFT(t, s) + b
	/* Do the following 16 operations. */
	S(a, b, c, d,  1,  5, T17);
	S(d, a, b, c,  6,  9, T18);
	S(c, d, a, b, 11, 14, T19);
	S(b, c, d, a,  0, 20, T20);
	S(a, b, c, d,  5,  5, T21);
	S(d, a, b, c, 10,  9, T22);
	S(c, d, a, b, 15, 14, T23);
	S(b, c, d, a,  4, 20, T24);
	S(a, b, c, d,  9,  5, T25);
	S(d, a, b, c, 14,  9, T26);
	S(c, d, a, b,  3, 14, T27);
	S(b, c, d, a,  8, 20, T28);
	S(a, b, c, d, 13,  5, T29);
	S(d, a, b, c,  2,  9, T30);
	S(c, d, a, b,  7, 14, T31);
	S(b, c, d, a, 12, 20, T32);
#undef S
#undef G

	 /* Round 3. */
	 /* Let [abcd k s t] denote the operation
          a = b + ((a + H(b,c,d) + X[k] + T[i]) <<< s). */
#define H(x, y, z) ((x) ^ (y) ^ (z))
#define S(a, b, c, d, k, s, Ti)\
	t = a + H(b,c,d) + X[k] + Ti;\
	a = ROTATE_LEFT(t, s) + b
	/* Do the following 16 operations. */
	S(a, b, c, d,  5,  4, T33);
	S(d, a, b, c,  8, 11, T34);
	S(c, d, a, b, 11, 16, T35);
	S(b, c, d, a, 14, 23, T36);
	S(a, b, c, d,  1,  4, T37);
	S(d, a, b, c,  4, 11, T38);
	S(c, d, a, b,  7, 16, T39);
	S(b, c, d, a, 10, 23, T40);
	S(a, b, c, d, 13,  4, T41);
	S(d, a, b, c,  0, 11, T42);
	S(c, d, a, b,  3, 16, T43);
	S(b, c, d, a,  6, 23, T44);
	S(a, b, c, d,  9,  4, T45);
	S(d, a, b, c, 12, 11, T46);
	S(c, d, a, b, 15, 16, T47);
	S(b, c, d, a,  2, 23, T48);
#undef S
#undef H

	 /* Round 4. */
	 /* Let [abcd k s t] denote the operation
          a = b + ((a + I(b,c,d) + X[k] + T[i]) <<< s). */
#define I(x, y, z) ((y) ^ ((x) | ~(z)))
#define S(a, b, c, d, k, s, Ti)\
	t = a + I(b,c,d) + X[k] + Ti;\
	a = ROTATE_LEFT(t, s) + b
	/* Do the following 16 operations. */
	S(a, b, c, d,  0,  6, T49);
	S(d, a, b, c,  7, 10, T50);
	S(c, d, a, b, 14, 15, T51);
	S(b, c, d, a,  5, 21, T52);
	S(a, b, c, d, 12,  6, T53);
	S(d, a, b, c,  3, 10, T54);
	S(c, d, a, b, 10, 15, T55);
	S(b, c, d, a,  1, 21, T56);
	S(a, b, c, d,  8,  6, T57);
	S(d, a, b, c, 15, 10, T58);
	S(c, d, a, b,  6, 15, T59);
	S(b, c, d, a, 13, 21, T60);
	S(a, b, c, d,  4,  6, T61);
	S(d, a, b, c, 11, 10, T62);
	S(c, d, a, b,  2, 15, T63);
	S(b, c, d, a,  9, 21, T64);
#undef S
#undef I

	/* Then perform the following additions. (That is increment each
	   of the four registers by the value it had before this block
	   was started.)
	 */
	pms->abcd[0] += a;
	pms->abcd[1] += b;
	pms->abcd[2] += c;
	pms->abcd[3] += d;
}

/*
   Independent implementation of MD4 (RFC 1320).

   This code implements the MD4 Algorithm defined in RFC 1320.
   It is derived directly from the text of the RFC and not from the
   reference implementation.

   md4_process is the original md5_process() function modified by Robert
   Osterlund for use with PIKT.
 */

void
md4_process(mdx_state_t *pms, const mdx_byte_t *data /*[64]*/)
{
	mdx_word_t
		a = pms->abcd[0], b = pms->abcd[1],
		c = pms->abcd[2], d = pms->abcd[3];
	mdx_word_t t;

#ifndef ARCH_IS_BIG_ENDIAN
#  define ARCH_IS_BIG_ENDIAN 1	/* slower, default implementation */
#endif

#if ARCH_IS_BIG_ENDIAN

	/*
	 * On big-endian machines, we must arrange the bytes in the right
	 * order.  (This also works on machines of unknown byte order.)
	 */
	mdx_word_t X[16];
	const mdx_byte_t *xp = data;
	int i;

	for (i=0; i<16; ++i, xp += 4)
		X[i] = xp[0] + (xp[1] << 8) + (xp[2] << 16) + (xp[3] << 24);

#else  /* ! ARCH_IS_BIG_ENDIAN */

	/*
	 * On little-endian machines, we can process properly aligned data
	 * without copying it.
	 */
	mdx_word_t xbuf[16];
	const mdx_word_t *X;

	if (! ((data - (const mdx_byte_t *)0) & 3)) {
		/* data are properly aligned */
		X = (const mdx_word_t *)data;
	} else {
		/* not aligned */
		memcpy(xbuf, data, 64);
		X = xbuf;
	}
#endif

#define ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32 - (n))))

	/* Round 1. */
	/* Let [abcd k s] denote the operation
	a = (a + F(b,c,d) + X[k]) <<< s. */
#define F(x, y, z) (((x) & (y)) | (~(x) & (z)))
#define S(a, b, c, d, k, s)\
	t = a + F(b,c,d) + X[k];\
	a = ROTATE_LEFT(t, s)
	/* Do the following 16 operations. */
	S(a, b, c, d,  0,  3);
	S(d, a, b, c,  1,  7);
	S(c, d, a, b,  2, 11);
	S(b, c, d, a,  3, 19);
	S(a, b, c, d,  4,  3);
	S(d, a, b, c,  5,  7);
	S(c, d, a, b,  6, 11);
	S(b, c, d, a,  7, 19);
	S(a, b, c, d,  8,  3);
	S(d, a, b, c,  9,  7);
	S(c, d, a, b, 10, 11);
	S(b, c, d, a, 11, 19);
	S(a, b, c, d, 12,  3);
	S(d, a, b, c, 13,  7);
	S(c, d, a, b, 14, 11);
	S(b, c, d, a, 15, 19);
#undef S
#undef F

	 /* Round 2. */
	 /* Let [abcd k s] denote the operation
          a = (a + G(b,c,d) + X[k] + 5A827999) <<< s). */
#define G(x, y, z) (((x) & (y)) | ((x) & (z)) | ((y) & (z)))
#define S(a, b, c, d, k, s)\
	t = a + G(b,c,d) + X[k] + 0x5a827999;\
	a = ROTATE_LEFT(t, s)
	/* Do the following 16 operations. */
	S(a, b, c, d,  0,  3);
	S(d, a, b, c,  4,  5);
	S(c, d, a, b,  8,  9);
	S(b, c, d, a, 12, 13);
	S(a, b, c, d,  1,  3);
	S(d, a, b, c,  5,  5);
	S(c, d, a, b,  9,  9);
	S(b, c, d, a, 13, 13);
	S(a, b, c, d,  2,  3);
	S(d, a, b, c,  6,  5);
	S(c, d, a, b, 10,  9);
	S(b, c, d, a, 14, 13);
	S(a, b, c, d,  3,  3);
	S(d, a, b, c,  7,  5);
	S(c, d, a, b, 11,  9);
	S(b, c, d, a, 15, 13);
#undef S
#undef G

	 /* Round 3. */
	 /* Let [abcd k s] denote the operation
          a = (a + H(b,c,d) + X[k] + 6ED9EBA1) <<< s). */
#define H(x, y, z) ((x) ^ (y) ^ (z))
#define S(a, b, c, d, k, s)\
	t = a + H(b,c,d) + X[k] + 0x6ed9eba1;\
	a = ROTATE_LEFT(t, s)
	/* Do the following 16 operations. */
	S(a, b, c, d,  0,  3);
	S(d, a, b, c,  8,  9);
	S(c, d, a, b,  4, 11);
	S(b, c, d, a, 12, 15);
	S(a, b, c, d,  2,  3);
	S(d, a, b, c, 10,  9);
	S(c, d, a, b,  6, 11);
	S(b, c, d, a, 14, 15);
	S(a, b, c, d,  1,  3);
	S(d, a, b, c,  9,  9);
	S(c, d, a, b,  5, 11);
	S(b, c, d, a, 13, 15);
	S(a, b, c, d,  3,  3);
	S(d, a, b, c, 11,  9);
	S(c, d, a, b,  7, 11);
	S(b, c, d, a, 15, 15);
#undef S
#undef H

	/* Then perform the following additions. (That is increment each
	   of the four registers by the value it had before this block
	   was started.)
	 */
	pms->abcd[0] += a;
	pms->abcd[1] += b;
	pms->abcd[2] += c;
	pms->abcd[3] += d;
}

/* calculate the MDx checksum, also the number of bytes, for the given file

   Written by Robert Osterlund for use with PIKT.
 */
char *
mdxsumfil (const char *file, int type)
{
	static char fnnam[] = "mdxsumfil";
	mdx_state_t s;
	mdx_byte_t d[16];
	FILE *fp;
	unsigned char buffer[1024];
	int bytes_read;
	long total_bytes;
	char *r;

	if ((fp = fopen(file, "rb")) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", file);
		r = mystrdup(errstr, fnnam);
		return r;
	}
	total_bytes = 0;
	mdx_init(&s, type);
	while ((bytes_read = fread(buffer, 1, 1024, fp))) {
		mdx_append(&s, buffer, bytes_read, type);
		total_bytes += bytes_read;
	}
	mdx_finish(&s, d, type);
	if (ferror(fp) || fclose(fp)) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "mdxsumfil(), getc() or fclose() error for %s", file);
		r = mystrdup(errstr, fnnam);
		return r;
	}
	r = mymalloc(strlen(file)+45+PADDING, fnnam);
	sprintf(r, "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x %ld %s", d[0], d[1], d[2], d[3], d[4], d[5], d[6], d[7], d[8], d[9], d[10], d[11], d[12], d[13], d[14], d[15], total_bytes, file);
	return r;
}

/* calculate the MDx checksum for the given string

   Written by Robert Osterlund for use with PIKT.
 */
char *
mdxsumstr (const char *str, int type)
{
	static char fnnam[] = "mdxsumstr";
	mdx_state_t s;
	mdx_byte_t d[16];
	unsigned char buffer[1024];
	int bytes_read;
	long i;
	char *r;

	mdx_init(&s, type);
	bytes_read = 0;
	i = 0;
	while ((buffer[bytes_read] = (unsigned char) *(str+i))) {
		bytes_read++;
		i++;
		if ((! *(str+i)) || (bytes_read == 1024)) {
			mdx_append(&s, buffer, bytes_read, type);
			bytes_read = 0;
		}
	}
	mdx_finish(&s, d, type);
	r = mymalloc(33, fnnam);
	sprintf(r, "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x", d[0], d[1], d[2], d[3], d[4], d[5], d[6], d[7], d[8], d[9], d[10], d[11], d[12], d[13], d[14], d[15]);
	return r;
}

/*****************************************************************************/
